#ifndef TSCCDEA_H_
#define TSCCDEA_H_

#include <vector>
#include <map>
#include <set>
#include "tsccd-puzzle.h"
#include "tsccd-symmap.h"
#include "twister.h"

extern MTRand mt;


class TSCCD_EA
{
	public:
		//////////////////////////////////////////////////////////////////////////
		struct Params
		{
			unsigned int tsccd_k, tsccd_v;
			
			unsigned int popsize;
			unsigned int offspring;	 // offspring to generate from each construction
			unsigned int maxinitlen; // maximum length of initial genomes
			
			struct Mutate
			{
				float mutaterate;				 // chance of inserting a random move instead of an indv.
			} mutate;
		};
		//////////////////////////////////////////////////////////////////////////
		struct Stats
		{
			unsigned long constructions;
			float avg_fit;
			float avg_len;				// average genome length
			unsigned int max_run; // length of longest move sequence encountered
														// which causes no errors
			float best_cover;			// cover ratio of best ever solution
			bool solved;
		};
		//////////////////////////////////////////////////////////////////////////
		
		TSCCD_EA ();
		~TSCCD_EA ();
		
		void initialize ( const Params & params );
			// set up new experiment with given parameters
		
		void breed ();
			// evaluate some fitnesses and create/queue offspring
		
		void select ();
			// do survival selection, clear offspring queue
		
		Stats getStats ();
		
		TSCCD_Puzzle * getBest ();
		
	private:
		class Individual;
		
		Individual * getIndvForCover ( const TSCCD_Puzzle & puz );
			// find an individual that has the most positive impact on the cover of puz
		
		Individual * getWorstIndv ();
			// find the worst-fitness individual
		
		void randRange ( unsigned int & starti, unsigned int & endi,
										 const unsigned int maxi, const unsigned int minlen );
			// find a random range in [0..maxi] at least minlen long
		
		//////////////////////////////////////////////////////////////////////////
		class Individual
		{
			public:
				Individual ( const unsigned int k, const unsigned int v );
				Individual ( const Individual & src );
				~Individual ();
				
				const Individual & operator= ( const Individual & rhs );
				
				bool operator< ( const Individual & rhs ) const;
					// compare lastuse
				
				void setLeadin ( const TSCCD_Block & newleadin );
				
				const TSCCD_Block & getLeadin () const;
				const std::vector< TSCCD_Move > & getMoves() const;
								
				void pushMove ( const TSCCD_Move & newmove );
					// add the given move to the end of the genome
				
				int selfScore () const;
					// return score of this piece alone
				
				void scramble ( const unsigned int length );
					// totally scramble genome, use 'length' moves
				
				bool mutate ( const TSCCD_EA::Params::Mutate & params );
					// returns true iff there was any change
				
				void setFitness ( const float fitness );
				void updateFitness ( const float fitness );
					// setF to set & reset running average
					// updateF to update a running average
				
				float fitness () const; // fitness is average score contribution per move
				
				void setLastUse ( const unsigned int iteration );
				unsigned int getLastUse () const;
					// for age/relevance-based selection
				
			private:
				const unsigned int m_k, m_v;
				
				// genome
				TSCCD_Block m_leadin;
				std::vector< TSCCD_Move > m_moves;
				
				float m_fitness;
				unsigned int m_fitnessupdates;
				unsigned int m_lastuse;
		};
		//////////////////////////////////////////////////////////////////////////
		
		std::vector< Individual > m_pop;
		
		TSCCD_Puzzle * m_bestever; // contains best total solution yet found

		Params m_params;
		Stats m_stats;
};


#endif /*TSCCDEA_H_*/
