
#include "tsccd-puzzle.h"


std::ostream & operator<< ( std::ostream & os, const TSCCD_Puzzle & puzzle )
{
	const size_t blocks = puzzle.m_blocks.size();
	
	// block contents
	for( size_t row = 0; row < puzzle.getK(); row++ )
	{
		for( size_t block = 0; block < blocks; block++ )
		{
			std::string symc( 1, (char)( 0x41 + puzzle.m_blocks[ block ][ row ] ) );
			if( block == 0 )
				os << symc << "  ";
			else
				if( puzzle.m_movehistory[ block - 1 ].pos() == row )
					os << "[" << symc << "] ";
				else
					os << " " << symc << "  ";
		}
		os << std::endl;
	}
	
	os << std::endl;
	for( size_t m = 0; m < blocks - 1; m++ )
	{
		os << "(" << puzzle.m_movehistory[ m ].pos() << ",";
		os << std::string( 1, 'A' + (char)puzzle.m_movehistory[ m ].symbol() ) << ") ";
	}
	os << std::endl;
	
	return os;
}


TSCCD_Puzzle::TSCCD_Puzzle ( const unsigned int k, const unsigned int v )
	: m_k( k ), m_v( v ),
		m_solutionlength( ( m_v*(m_v-1) - m_k*(m_k-1) ) / 2 / ( m_k - 1 ) ),
		m_cover( v )
{
	reset();
}


TSCCD_Puzzle::TSCCD_Puzzle ( const TSCCD_Puzzle & src )
	: m_k( src.m_k ), m_v( src.m_v ), m_solutionlength( src.m_solutionlength ),
		m_movehistory( src.m_movehistory ), m_blocks( src.m_blocks ),
		m_cover( src.m_cover ), m_covermisses( src.m_covermisses )
{}


TSCCD_Puzzle::TSCCD_Puzzle ( const TSCCD_Puzzle & base, const TSCCD_Move & applymove )
	: m_k( base.m_k ), m_v( base.m_v ), m_solutionlength( base.m_solutionlength ),
		m_cover( base.m_cover )
{
	m_movehistory = base.m_movehistory;
	m_blocks = base.m_blocks;
	m_covermisses = base.m_covermisses;
	applyMove( applymove );
}


const TSCCD_Puzzle & TSCCD_Puzzle::operator= ( const TSCCD_Puzzle & rhs )
{
	if( this == &rhs )
		return rhs;
	if( m_k != rhs.m_k || m_v != rhs.m_v )
		throw string( "puzzle bad =" );
	
	m_movehistory = rhs.m_movehistory;
	m_blocks = rhs.m_blocks;
	m_cover = rhs.m_cover;
	m_covermisses = rhs.m_covermisses;
	return *this;
}


const unsigned int & TSCCD_Puzzle::getK () const { return m_k; }

const unsigned int & TSCCD_Puzzle::getV () const { return m_v; }

const std::vector< TSCCD_Block > & TSCCD_Puzzle::getBlocks () const { return m_blocks; }

const std::vector< TSCCD_Move > & TSCCD_Puzzle::getMoves () const { return m_movehistory; }


bool TSCCD_Puzzle::hasASolution () const
{
	return ( ( m_v*(m_v-1) - m_k*(m_k-1) ) / 2 ) % ( m_k - 1 ) == 0;
}


unsigned int TSCCD_Puzzle::solutionLength () const { return m_solutionlength; }


unsigned int TSCCD_Puzzle::movesLeft () const
{
	if( m_movehistory.size() >= m_solutionlength )
		return 0;
	else
		return m_solutionlength - m_movehistory.size();
}


bool TSCCD_Puzzle::isSolved () const
{
	return ( coveredPairs() == totalPairs() && misses() == 0 );
}


unsigned int TSCCD_Puzzle::totalPairs () const { return m_cover.pairsTotal(); }

unsigned int TSCCD_Puzzle::coveredPairs () const { return m_cover.pairsCovered(); }

unsigned int TSCCD_Puzzle::misses () const { return m_covermisses; }


TSCCD_Puzzle TSCCD_Puzzle::normalize ()
{
	TSCCD_Puzzle mappedpuz( m_k, m_v );
	
	TSCCD_SymMap map( m_v );
	map.captureMap( *m_blocks.begin(), mappedpuz.getBlocks().back() );
	
	for( unsigned int i = 0; i < m_movehistory.size(); i++ )
		mappedpuz.applyMove( map.mapMove( m_movehistory[ i ] ) );
	
	return mappedpuz;
}


void TSCCD_Puzzle::reset ()
{
	m_movehistory.clear();
	m_movehistory.reserve( solutionLength() );
	
	m_blocks.clear();
	m_blocks.reserve( solutionLength() + 1);
	std::vector<unsigned int> firstblockvals;
	for( unsigned int i = 0; i < m_k; i++ )
		firstblockvals.push_back( i );
	m_blocks.push_back( TSCCD_Block( firstblockvals ) );
	
	m_cover.reset();
	for( unsigned int i = 0; i < m_k - 1; i++ )
		for( unsigned int j = i + 1; j < m_k; j++ )
			m_cover.cover( i, j );
	m_covermisses = 0;
}


void TSCCD_Puzzle::reset ( const TSCCD_Block & startblock )
{
	m_movehistory.clear();
	
	m_blocks.clear();
	m_blocks.push_back( startblock );
	
	m_cover.reset();
	for( unsigned int i = 0; i < m_k - 1; i++ )
		for( unsigned int j = i + 1; j < m_k; j++ )
			m_cover.cover( startblock[i], startblock[j] );
	m_covermisses = 0;	
}


int TSCCD_Puzzle::scoreMoves ( const std::vector< TSCCD_Move > & moves,
															 const TSCCD_SymMap & mapping ) const
{
	TSCCD_Puzzle testpuz( *this );
	
	for( unsigned int i = 0; i < moves.size(); i++ )
		testpuz.applyMove( mapping.mapMove( moves[ i ] ) );
	
	return (int)( testpuz.coveredPairs() - this->coveredPairs() ) -
				 (int)( testpuz.misses() - this->misses() );
}


int TSCCD_Puzzle::applyMove ( const TSCCD_Move & newmove )
{
	int score = 0;
	m_movehistory.push_back( newmove );
	
	m_blocks.push_back( m_blocks.back() );
	m_blocks.back().applyMove( newmove );
	
	for( unsigned int p = 0; p < m_k; p++ )
		if( p != newmove.pos() )
			if( m_cover.cover( newmove.symbol(), m_blocks.back()[ p ] ) )
				score++;
			else
			{
				m_covermisses++;
				score--;
			}
	
	return score;
}


std::vector< TSCCD_Move > TSCCD_Puzzle::getValidMoves ()
{
	std::vector< TSCCD_Move > moves;
	std::vector< bool > valmask( m_k, true );
	for( unsigned int i = 0; i < m_k; i++ )
		valmask[ m_blocks.back()[i] == false ];
	for( unsigned int v = 0; v < m_v; v++ )
		if( valmask[v] )
			for( unsigned int pos = 0; pos < m_k; pos++ )
				moves.push_back( TSCCD_Move( pos, v ) );
	return moves;
}


std::vector< TSCCD_Move > TSCCD_Puzzle::getPossibleMoves ()
{
	std::vector< TSCCD_Move > moves;
	for( unsigned int pos = 0; pos < m_k; pos++ )
		for( unsigned int v = 0; v < m_v; v++ )
			moves.push_back( TSCCD_Move( pos, v ) );
	return moves;
}

