class TSCCD_Puzzle;

#ifndef TSCCDPUZZLE_H_
#define TSCCDPUZZLE_H_

#include <vector>
#include <iostream>
#include <iomanip>
#include <string>
#include "paircover.h"
#include "tsccd-move.h"
#include "tsccd-block.h"
#include "tsccd-symmap.h"


std::ostream & operator<< ( std::ostream & os, const TSCCD_Puzzle & puzzle );
// formatted output of current puzzle state


class TSCCD_Puzzle
{
	friend std::ostream & operator<< ( std::ostream& os, const TSCCD_Puzzle & puzzle );
	public:
		TSCCD_Puzzle ( const unsigned int k, const unsigned int v );
			// k: size of each block
			// v: number of symbols
		
		TSCCD_Puzzle ( const TSCCD_Puzzle & src );
		
		TSCCD_Puzzle ( const TSCCD_Puzzle & base, const TSCCD_Move & applymove );
			// copy base and apply the move
		
		const TSCCD_Puzzle & operator= ( const TSCCD_Puzzle & rhs );
		
		const unsigned int & getK () const;
		const unsigned int & getV () const;
		const std::vector< TSCCD_Block > & getBlocks () const;
		const std::vector< TSCCD_Move > & getMoves () const;
		
		bool hasASolution () const;
			// returns true iff this puzzle is solvable
			// NOTE: returns some false positives
		
		unsigned int solutionLength () const;
			// return total number of moves required to solve this puzzle
		
		unsigned int movesLeft () const;
			// return number of moves remaining
		
		bool isSolved () const;
			// returns true iff current state is a solution
		
		unsigned int totalPairs () const;
			// return total number of pairs in this puzzle
		
		unsigned int coveredPairs () const;
			// return number of pairs covered at this state
		
		unsigned int misses () const;
			// return number of incorrect pairs generated at this state
		
		TSCCD_Puzzle normalize ();
			// remap the puzzle such that the initial block is [0..k-1]
		
		void reset ();
			// return puzzle to initial state, with single initial block [0..k-1]
		
		void reset ( const TSCCD_Block & startblock );
			// return puzzle to initial state, using given starting block
		
		int scoreMoves ( const std::vector< TSCCD_Move > & moves,
										 const TSCCD_SymMap & mapping ) const;
			// predict the result of applying the given moves, remapped, in order
			// +1 for every new cover, -1 for every re-cover
		
		int applyMove ( const TSCCD_Move & newmove );
			// add a block to the puzzle, according to newmove
			// return ( new covers - errors caused )
		
		std::vector< TSCCD_Move > getValidMoves ();
			// return all plausible moves for next block
			// this does not consider hits/misses, only obviously invalid moves
		
		std::vector< TSCCD_Move > getPossibleMoves ();
			// return all moves conceivable for next block, valid or not
						
	private:
		const unsigned int m_k;
		const unsigned int m_v;
		
		const unsigned int m_solutionlength;
		
		std::vector< TSCCD_Move > m_movehistory;
		std::vector< TSCCD_Block > m_blocks;
			// movehistory and blocks indices are not synchronized, as block 0 has no move
		
		PairCover m_cover;
		unsigned int m_covermisses;
};

#endif /*TSCCDPUZZLE_H_*/
